﻿/*
 * Készítette a SharpDevelop.
 * Felhasználó: phil
 * Dátum: 2009.11.20.
 * Idő: 13:32
 * 
 * A sablon megváltoztatásához használja az Eszközök | Beállítások | Kódolás | Szabvány Fejlécek Szerkesztését.
 */
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Windows.Forms;
using System.Windows.Forms.DataVisualization;
using System.Windows.Forms.DataVisualization.Charting;
//
using BSE.Windows.Forms;
//
using Noise.Controls.MSChart;
using Noise.DataAcquisition;
using TetheredSun.Controls;
using TetheredSun.Core;
using TetheredSun.SignalAnalysis;

namespace GenericMonitor
{
	/// <summary>
	/// Description of MainForm.
	/// </summary>
	public partial class MainForm : Form
	{
		#region FIELDS
		protected Ftdi ftdi;
		protected Edaq530 device;
		protected ChannelInfo channelAInfo = new ChannelInfo("A");
		protected ChannelInfo channelBInfo = new ChannelInfo("B");
		protected ChannelInfo channelCInfo = new ChannelInfo("C");
		protected ChannelDataPipe channelA = new ChannelDataPipe();
		protected ChannelDataPipe channelB = new ChannelDataPipe();
		protected ChannelDataPipe channelC = new ChannelDataPipe();
		protected ChannelData channelAPlot;	
		protected ChannelData channelBPlot;
		protected ChannelData channelCPlot;
		protected ChannelData channelABuffer;	
		protected ChannelData channelBBuffer;
		protected ChannelData channelCBuffer;
		protected double dt;
		protected bool isSampling = false;
		protected bool textBoxesInitialised = false;
		protected int defaultCycles = 8;
		protected int dataBufferRefreshInterval = 100;
		protected int maxPoints = 10000;
		protected string meterFormat = "f2";
		protected Measurement measurement = new Measurement();
		protected DataBuffer dataBuffer = new DataBuffer();
		protected System.Windows.Forms.Timer meterRefreshTimer = new System.Windows.Forms.Timer();
		#endregion
		
		#region CONSTRUCTORS
		public MainForm()
		{
			//
			// The InitializeComponent() call is required for Windows Forms designer support.
			//
			InitializeComponent();
			SetFtdi();
			InitialiseDevice();
			InitialiseForm();
			LoadMeasurement("LastSettings.xml");
			SetControls();
			GetVersion();
		}
		#endregion
		
		#region METHODS
		public void AddPanelExpandingEventHandlers()
		{
			foreach (XPanderPanel panel in xPanderPanelList1.XPanderPanels) {
				panel.PanelExpanding += HandlePanelExpandingEvents;
			}
		}
		
		public void CopyToClipBoard(DataGridView dataGridView)
		{
			string clipboardString = String.Empty;
			
			// Write the header:
			if (dataGridView == aDataGridView) {
				clipboardString += "Channel A\t";
				if (measurement[0].Sensor != null) {
					clipboardString += measurement[0].Sensor.GetLabel();
				} else {
					clipboardString += "U [V]";
				}
			} else if (dataGridView == bDataGridView) {
				clipboardString += "Channel B\t";
				if (measurement[1].Sensor != null) {
					clipboardString += measurement[1].Sensor.GetLabel();
				} else {
					clipboardString += "U [V]";
				}
			} else if (dataGridView == cDataGridView) {
				clipboardString += "Channel C\t";
				if (measurement[2].Sensor != null) {
					clipboardString += measurement[2].Sensor.GetLabel();
				} else {
					clipboardString += "U [V]";
				}
			}
			
			clipboardString += "\t" + Environment.NewLine;
			
			for (int i = 0; i < dataGridView.ColumnCount; i++) {
				clipboardString += dataGridView.Columns[i].HeaderText + "\t";
			}
			
			clipboardString = clipboardString.Trim("\t".ToCharArray());
			clipboardString += Environment.NewLine;
			
			// Write the rest of the data:
			for (int i = 0; i < dataGridView.RowCount - 1; i++) {
				for (int j = 0; j < dataGridView.ColumnCount; j++) {
					clipboardString += (dataGridView.Rows[i].Cells[j].Value != null) ?
						dataGridView.Rows[i].Cells[j].Value.ToString() : String.Empty;
					clipboardString += "\t";
				}
				clipboardString = clipboardString.Trim("\t".ToCharArray());
				clipboardString += Environment.NewLine;
			}
			

	
			Clipboard.SetDataObject(clipboardString);
		}
		
		public void GetSamplingParameters()
		{
			double? fs = device.SamplingFrequency;
			
			measurement.Averages = device.Averages;
			if (measurement.Averages == null) measurement.Averages = Averages.One;
			measurement.SamplingFrequency = fs.HasValue ? fs.Value : 100.0;
			dt = 1.0 / measurement.SamplingFrequency;
			//
			fsNumericControl.SetValue(measurement.SamplingFrequency);
			//
			averagesComboBox.SelectedIndex = measurement.Averages.Code;
			//
			xWidthNumericControl.SetValue(measurement.XScalePoints * dt);
		}
		
		public void GetVersion()
		{
			AssemblyName assemblyName = Assembly.GetExecutingAssembly().GetName();
			
			versionLabel.Text = String.Format("Version: {0}.{1}{2}{3}", assemblyName.Version.Major, assemblyName.Version.Minor,
			                                 assemblyName.Version.Build, assemblyName.Version.Revision);
		}
		
		public void InitialiseDevice()
		{
			if (ftdi == null) {
				device = null;
				fsNumericControl.Enabled = false;
				averagesComboBox.Enabled = false;
				startToolStripButton.Enabled = false;
				return;
			}
			
			
			device = new Edaq530(ftdi);
			device.Reset();
			device.Channels.Add(channelA);
			device.Channels.Add(channelB);
			device.Channels.Add(channelC);
			//
			toolStripStatusLabel1.Text = device.GetDeviceInfo();
			//
			fsNumericControl.Enabled = true;
			averagesComboBox.Enabled = true;
			startToolStripButton.Enabled = true;
		}
				
		public void InitialiseForm()
		{
			AddPanelExpandingEventHandlers();
			
			aSensorControl.SensorRecreated += HandleSensorRecreatedEvents;
			aSensorControl.SensorUpdated += HandleSensorUpdatedEvents;
			bSensorControl.SensorRecreated += HandleSensorRecreatedEvents;
			bSensorControl.SensorUpdated += HandleSensorUpdatedEvents;
			cSensorControl.SensorRecreated += HandleSensorRecreatedEvents;
			cSensorControl.SensorUpdated += HandleSensorUpdatedEvents;
			//
			xPointsNumericControl.MaxValue = maxPoints;
			//
			meterRefreshTimer.Interval = 500;
			meterRefreshTimer.Tick += RefreshMeters;
			//
			measurement[0].Colour = Color.DarkCyan;
			measurement[1].Colour = Color.Maroon;
			measurement[2].Colour = Color.RoyalBlue;
			//
			aDataPlotter.Initialise();
			bDataPlotter.Initialise();
			cDataPlotter.Initialise();
		}
		
		public void LoadMeasurement(string path)
		{
			try {
				measurement = Measurement.Open(path);
			} catch (Exception exception) {
				ExceptionReporter.Show(exception, true);
				measurement = new Measurement();
				measurement[0].Colour = Color.DarkCyan;
				measurement[1].Colour = Color.Maroon;
				measurement[2].Colour = Color.RoyalBlue;
			} finally {
				SetControls();
			}
		}
		
		public void RefreshMeters(object sender, EventArgs e)
		{
			if (sender == meterRefreshTimer) {
				if (channelAPlot != null) {
					aTextBox.Text = (channelAPlot.ChannelInfo.Sensor != null) ?
						channelAPlot.ChannelInfo.Sensor.GetValueString(channelAPlot.CurrentValue, meterFormat, true) :
						String.Format("U = {0} V", channelAPlot.CurrentValue.ToString(meterFormat));
					                                                                                                           
				} else {
					aTextBox.Text = "NO DATA";
				}
				//
				if (channelBPlot != null) {
					bTextBox.Text = (channelBPlot.ChannelInfo.Sensor != null) ?
						channelBPlot.ChannelInfo.Sensor.GetValueString(channelBPlot.CurrentValue, meterFormat, true) :
						String.Format("U = {0} V", channelBPlot.CurrentValue.ToString(meterFormat));
					                                                                                                           
				} else {
					bTextBox.Text = "NO DATA";
				}
				//
				if (channelCPlot != null) {
					cTextBox.Text = (channelCPlot.ChannelInfo.Sensor != null) ?
						channelCPlot.ChannelInfo.Sensor.GetValueString(channelCPlot.CurrentValue, meterFormat, true) :
						String.Format("U = {0} V", channelCPlot.CurrentValue.ToString(meterFormat));
					                                                                                                           
				} else {
					cTextBox.Text = "NO DATA";
				}
				
				if (!textBoxesInitialised) {
					SetTextBoxes();
					textBoxesInitialised = true;
				}
			}
		}
		
		public void Reset(bool resetDevice)
		{
			if (resetDevice && (device != null)) device.Reset();
			// Reset queues:
			lock (channelABuffer.SyncRoot) { channelABuffer.Clear(); }
			lock (channelAPlot.SyncRoot) { channelAPlot.Clear(); }
			lock (channelBBuffer.SyncRoot) { channelBBuffer.Clear(); }
			lock (channelBPlot.SyncRoot) { channelBPlot.Clear(); }
			lock (channelCBuffer.SyncRoot) { channelCBuffer.Clear(); }
			lock (channelCPlot.SyncRoot) { channelCPlot.Clear(); }
			//
			aDataPlotter.Reset();
			bDataPlotter.Reset();
			cDataPlotter.Reset();
			dataBuffer.Clear();
			UpdateDataGridView(aDataGridView);
			UpdateDataGridView(bDataGridView);
			UpdateDataGridView(cDataGridView);
			aTextBox.Text = String.Empty;
			bTextBox.Text = String.Empty;
			cTextBox.Text = String.Empty;
		}
		
		public void SetControls()
		{
			// Generic:
			dt = 1.0 / measurement.SamplingFrequency;
			
			fsNumericControl.SetValue(measurement.SamplingFrequency);
			averagesComboBox.SelectedIndex = measurement.Averages.Code;
			xScaleTitleLabelledTextBox.TextBox.Text = measurement.XScaleTitle;
			xPointsNumericControl.SetValue(measurement.XScalePoints);
			xWidthNumericControl.SetValue(dt * measurement.XScalePoints);
			dataBufferNumericControl.SetValue(measurement.DataBufferSize);
			refreshRateNumericControl.SetValue(1000.0 / measurement.RefreshInterval);
			// Channel A:
			channelAInfo.Sensor = measurement[0].Sensor;
			aSensorControl.Sensor = measurement[0].Sensor;
			HandleSensorUpdatedEvents(aSensorControl, new EventArgs());
			aActiveCheckBox.Checked = measurement[0].IsActive;
			aAutoCheckBox.Checked = measurement[0].IsAutoscaled;
			if (aAutoCheckBox.Checked) {
				aMinNumericControl.Enabled = false;
				aMinNumericControl.SetValue(Double.NaN, "Auto");
				aMaxNumericControl.Enabled = false;
				aMaxNumericControl.SetValue(Double.NaN, "Auto");
				aDataPlotter.ChartAreas[0].AxisY.Minimum = Double.NaN;
				aDataPlotter.ChartAreas[0].AxisY.Maximum = Double.NaN;
				aDataPlotter.ChartAreas[0].AxisY.IsStartedFromZero = false;
			} else {
				aMinNumericControl.Enabled = true;
				aMinNumericControl.SetValue(measurement[0].YMin);
				aMaxNumericControl.Enabled = true;
				aMaxNumericControl.SetValue(measurement[0].YMax);
				aDataPlotter.ChartAreas[0].AxisY.Minimum = measurement[0].YMin;
				aDataPlotter.ChartAreas[0].AxisY.Maximum = measurement[0].YMax;
			}
			if (measurement[0].LevelCrossingDetector != null) {
				aLevelCrossingControl.Set(measurement[0].LevelCrossingDetector, measurement[0].ObjectLength);
				aDataPlotter.LevelCrossingDetector = measurement[0].LevelCrossingDetector;
				aDataPlotter.LevelCrossingDetector.Cycles.Capacity = defaultCycles;
				aDataPlotter.LevelCrossingDetector.CyclesUpdated += HandleCyclesUpdatedEvents;
			}
			aTextBox.ForeColor = measurement[0].Colour;
			aColourPickerControl.Colour = measurement[0].Colour;
			aDataPlotter.Series[0].Color = measurement[0].Colour;
			aDataPlotter.ChartAreas[0].AxisX.Title = measurement.XScaleTitle;
			// Channel B:
			channelBInfo.Sensor = measurement[1].Sensor;
			bSensorControl.Sensor = measurement[1].Sensor;
			HandleSensorUpdatedEvents(bSensorControl, new EventArgs());
			bActiveCheckBox.Checked = measurement[1].IsActive;
			bAutoCheckBox.Checked = measurement[1].IsAutoscaled;
			if (bAutoCheckBox.Checked) {
				bMinNumericControl.Enabled = false;
				bMinNumericControl.SetValue(Double.NaN, "Auto");
				bMaxNumericControl.Enabled = false;
				bMaxNumericControl.SetValue(Double.NaN, "Auto");
				bDataPlotter.ChartAreas[0].AxisY.Minimum = Double.NaN;
				bDataPlotter.ChartAreas[0].AxisY.Maximum = Double.NaN;
				bDataPlotter.ChartAreas[0].AxisY.IsStartedFromZero = false;
			} else {
				bMinNumericControl.Enabled = true;
				bMinNumericControl.SetValue(measurement[1].YMin);
				bMaxNumericControl.Enabled = true;
				bMaxNumericControl.SetValue(measurement[1].YMax);
				bDataPlotter.ChartAreas[0].AxisY.Minimum = measurement[1].YMin;
				bDataPlotter.ChartAreas[0].AxisY.Maximum = measurement[1].YMax;
			}
			if (measurement[1].LevelCrossingDetector != null) {
				bLevelCrossingControl.Set(measurement[1].LevelCrossingDetector, measurement[1].ObjectLength);
				bDataPlotter.LevelCrossingDetector = measurement[1].LevelCrossingDetector;
				bDataPlotter.LevelCrossingDetector.Cycles.Capacity = defaultCycles;
				bDataPlotter.LevelCrossingDetector.CyclesUpdated += HandleCyclesUpdatedEvents;
			}
			bTextBox.ForeColor = measurement[1].Colour;
			bColourPickerControl.Colour = measurement[1].Colour;
			bDataPlotter.Series[0].Color = measurement[1].Colour;
			bDataPlotter.ChartAreas[0].AxisX.Title = measurement.XScaleTitle;
			// Channel C:
			channelCInfo.Sensor = measurement[2].Sensor;
			cSensorControl.Sensor = measurement[2].Sensor;
			HandleSensorUpdatedEvents(cSensorControl, new EventArgs());
			cActiveCheckBox.Checked = measurement[2].IsActive;
			cAutoCheckBox.Checked =  measurement[2].IsAutoscaled;
			if (cAutoCheckBox.Checked) {
				cMinNumericControl.Enabled = false;
				cMinNumericControl.SetValue(Double.NaN, "Auto");
				cMaxNumericControl.Enabled = false;
				cMaxNumericControl.SetValue(Double.NaN, "Auto");
				cDataPlotter.ChartAreas[0].AxisY.Minimum = Double.NaN;
				cDataPlotter.ChartAreas[0].AxisY.Maximum = Double.NaN;
				cDataPlotter.ChartAreas[0].AxisY.IsStartedFromZero = false;
			} else {
				cMinNumericControl.Enabled = true;
				cMinNumericControl.SetValue(measurement[2].YMin);
				cMaxNumericControl.Enabled = true;
				cMaxNumericControl.SetValue(measurement[2].YMax);
				cDataPlotter.ChartAreas[0].AxisY.Minimum = measurement[2].YMin;
				cDataPlotter.ChartAreas[0].AxisY.Maximum = measurement[2].YMax;
			}
			if (measurement[2].LevelCrossingDetector != null) {
				cLevelCrossingControl.Set(measurement[2].LevelCrossingDetector, measurement[2].ObjectLength);
				cDataPlotter.LevelCrossingDetector = measurement[2].LevelCrossingDetector;
				cDataPlotter.LevelCrossingDetector.Cycles.Capacity = defaultCycles;
				cDataPlotter.LevelCrossingDetector.CyclesUpdated += HandleCyclesUpdatedEvents;
			}
			cTextBox.ForeColor = measurement[2].Colour;
			cColourPickerControl.Colour = measurement[2].Colour;
			cDataPlotter.Series[0].Color = measurement[2].Colour;
			cDataPlotter.ChartAreas[0].AxisX.Title = measurement.XScaleTitle;
			SetPlotters();
			SetDataBuffer();
			SetChannelVisibility();
			SetTextBoxes();
			SetTableVisibility();
		}
		
		public void SetChannelVisibility()
		{
			TableLayoutRowStyleCollection rows1 = tableLayoutPanel1.RowStyles;
			TableLayoutRowStyleCollection rows2 = tableLayoutPanel2.RowStyles;
			LevelCrossingControl levelCrossingControl = null;
			MSDataPlotter dataPlotter = null;
			System.Windows.Forms.Panel panel = null;
			int validChannelCount = measurement.GetActiveChannelCount();
			float percentage;
			
			
			percentage = (validChannelCount > 0) ? 100.0f / validChannelCount : 0.0f;
			
			if (validChannelCount == 0) {
				tableLayoutPanel1.Visible = false;
				tableLayoutPanel2.Visible = false;
				aLevelCrossingControl.Enabled = false;
				bLevelCrossingControl.Enabled = false;
				cLevelCrossingControl.Enabled = false;
				if (aDataPlotter.LevelCrossingDetector != null) aDataPlotter.LevelCrossingDetector.IsActive = false;
				if (bDataPlotter.LevelCrossingDetector != null) bDataPlotter.LevelCrossingDetector.IsActive = false;
				if (cDataPlotter.LevelCrossingDetector != null) cDataPlotter.LevelCrossingDetector.IsActive = false;
			} else {
				tableLayoutPanel1.Visible = true;
				tableLayoutPanel2.Visible = true;
				for (int i = 0; i < 3; i++) {
					switch (i) {
						case 1:
							levelCrossingControl = bLevelCrossingControl;
							panel = bDataGridPanel;
							dataPlotter = bDataPlotter;
							break;
						case 2:
							levelCrossingControl = cLevelCrossingControl;
							panel = cDataGridPanel;
							dataPlotter = cDataPlotter;
							break;
						case 0:
						default:
							levelCrossingControl = aLevelCrossingControl;
							panel = aDataGridPanel;
							dataPlotter = aDataPlotter;
							break;
					}
					if (measurement[i].IsActive) {
						dataPlotter.Visible = true;
						panel.Visible = true;
						levelCrossingControl.Enabled = true;
						//
						rows1[i].SizeType = SizeType.Percent;
						rows1[i].Height = percentage;
						rows2[i].SizeType = SizeType.Percent;
						rows2[i].Height = percentage;
					} else {
						dataPlotter.Visible = false;
						if (dataPlotter.LevelCrossingDetector != null) dataPlotter.LevelCrossingDetector.IsActive = false;
						panel.Visible = false;
						levelCrossingControl.Enabled = false;
						//
						rows1[i].SizeType = SizeType.Absolute;
						rows1[i].Height = 0.0F;
						rows2[i].SizeType = SizeType.Absolute;
						rows2[i].Height = 0.0F;
					}
				}	
			}
		}
		
		public void SetDataBuffer()
		{
			if (channelABuffer == null) {
				channelABuffer = new ChannelData(channelAInfo);
				channelA.Add(channelABuffer);
				dataBuffer.Channels.Add(channelABuffer);
			}
			
			if (channelBBuffer == null) {
				channelBBuffer = new ChannelData(channelBInfo);
				channelB.Add(channelBBuffer);
				dataBuffer.Channels.Add(channelBBuffer);
			}
			
			if (channelCBuffer == null) {
				channelCBuffer = new ChannelData(channelCInfo);
				channelC.Add(channelCBuffer);
				dataBuffer.Channels.Add(channelCBuffer);
			}
			
			dataBuffer.Interval = dataBufferRefreshInterval;
		}
		
		public void SetFtdi()
		{
			string[] ftdiDevices = Ftdi.GetSerialNumbers();
			
			if (ftdi != null) {
				ftdi.Close();
			}
			
			if ((ftdiDevices != null) && (ftdiDevices.Length > 0) && !String.IsNullOrEmpty(ftdiDevices[0])) {
				toolStripStatusLabel1.Text = ftdiDevices[0];
				ftdi = new Ftdi(ftdiDevices[0]);
				ftdi.Open();	
				ftdi.BaudRate = Edaq530.BaudRate;
				ftdi.Handshake = Edaq530.Handshake;
				//ftdi.SetLatency((byte)2);
				ftdi.Initialise();
			} else {
				toolStripStatusLabel1.Text = "No Edaq530 device found.";
				MessageBox.Show(
					"No Edaq530 device has been found. Please connect the device, wait about 10 seconds, then select the \"Rescan devices\" option from" +
					" the \"Connexion\" menu or press CTRL + R.", "No Edaq530 device found");
				ftdi = null;
			}
		}
		
		public void SetPlotters()
		{
			if (channelAPlot == null) {
				channelAPlot = new ChannelData(channelAInfo);
				channelA.Add(channelAPlot);
				aDataPlotter.Channel = channelAPlot;
			}
			
			if (channelBPlot == null) {
				channelBPlot = new ChannelData(channelBInfo);
				channelB.Add(channelBPlot);
				bDataPlotter.Channel = channelBPlot;
			}
			
			if (channelCPlot == null) {
				channelCPlot = new ChannelData(channelCInfo);
				channelC.Add(channelCPlot);
				cDataPlotter.Channel = channelCPlot;
			}
			
			aDataPlotter.Interval = measurement.RefreshInterval;
			aDataPlotter.HistoryLength = measurement.XScalePoints;
			aDataPlotter.TimeIncrement = dt;
			aDataPlotter.Series[0].Color = measurement[0].Colour;
			//
			bDataPlotter.Interval = measurement.RefreshInterval;
			bDataPlotter.HistoryLength = measurement.XScalePoints;
			bDataPlotter.TimeIncrement = dt;
			bDataPlotter.Series[0].Color = measurement[1].Colour;
			//
			cDataPlotter.Interval = measurement.RefreshInterval;
			cDataPlotter.HistoryLength = measurement.XScalePoints;
			cDataPlotter.TimeIncrement = dt;
			cDataPlotter.Series[0].Color = measurement[2].Colour;
		}
		
		public void SetTableVisibility()
		{
			bool isVisible = aLevelCrossingControl.IsActive || bLevelCrossingControl.IsActive || cLevelCrossingControl.IsActive;
			float columnWidth = 245.0f;
			TableLayoutColumnStyleCollection columns = tableLayoutPanel1.ColumnStyles;
			
			columns[0].SizeType = SizeType.Percent;
			columns[1].SizeType = SizeType.Absolute;
			
			columns[1].Width = isVisible ? columnWidth : 0.0f;
			columns[0].Width = 100.0f;
		}
		
		public void SetTextBoxes()
		{
			tableLayoutPanel2.Enabled = false;
			//
			int rowCount = measurement.GetActiveChannelCount();
			int rowHeight;
			int rowWidth;
			int margin;
			int textBoxHeight;
			int oldMargin = aTextBox.Margin.Top;
			double heightFactor = Double.MaxValue;
			double widthFactor = Double.MaxValue;
			float resizeFactor;
			double temporary;
			Point location;
			Size size;
			Font oldFont;
			
			if (rowCount == 0) return;
			
			rowHeight = tableLayoutPanel2.Height / rowCount;
			rowWidth = tableLayoutPanel2.Width;
			
			if (measurement[0].IsActive) {
				// Height:
				temporary = (double)(rowHeight - 2 * oldMargin)/aTextBox.PreferredHeight;
				if (temporary < heightFactor) heightFactor = temporary;
				// Width:
				temporary = (double)(rowWidth - 2 * oldMargin)/aTextBox.PreferredSize.Width;
				if (temporary < widthFactor) widthFactor = temporary;
			}
			
			if (measurement[1].IsActive) {
				// Height:
				temporary = (double)(rowHeight - 2 * oldMargin)/bTextBox.PreferredHeight;
				if (temporary < heightFactor) heightFactor = temporary;
				// Width:
				temporary = (double)(rowWidth - 2 * oldMargin)/bTextBox.PreferredSize.Width;
				if (temporary < widthFactor) widthFactor = temporary;
			}
			
			if (measurement[2].IsActive) {
				// Height:
				temporary = (double)(rowHeight - 2 * oldMargin)/cTextBox.PreferredHeight;
				if (temporary < heightFactor) heightFactor = temporary;
				// Width:
				temporary = (double)(rowWidth - 2 * oldMargin)/cTextBox.PreferredSize.Width;
				if (temporary < widthFactor) widthFactor = temporary;
			}
			
			resizeFactor = (float)Math.Min(heightFactor, widthFactor);
			
			oldFont = aTextBox.Font;
			
			aTextBox.Font = new Font(oldFont.FontFamily, oldFont.Size * resizeFactor);
			bTextBox.Font = new Font(oldFont.FontFamily, oldFont.Size * resizeFactor);
			cTextBox.Font = new Font(oldFont.FontFamily, oldFont.Size * resizeFactor);
			
			
			textBoxHeight = aTextBox.Height;
			margin = (rowHeight - textBoxHeight) / 2;
			if (margin < 3) margin = 3;
			
			if (measurement[0].IsActive) {
				aTextBox.Visible = true;
				location = aTextBox.Location;
				size = aTextBox.Size;
				aTextBox.SetBounds(location.X, margin, size.Width, size.Height);
				margin += rowHeight;
			} else {
				aTextBox.Visible = false;
			}
			
			if (measurement[1].IsActive) {
				bTextBox.Visible = true;
				location = bTextBox.Location;
				size = bTextBox.Size;
				bTextBox.SetBounds(location.X, margin, size.Width, size.Height);
				margin += rowHeight;
			} else {
				bTextBox.Visible = false;
			}
			
			if (measurement[2].IsActive) {
				cTextBox.Visible = true;
				location = cTextBox.Location;
				size = cTextBox.Size;
				cTextBox.SetBounds(location.X, margin, size.Width, size.Height);
				margin += rowHeight;
			} else {
				cTextBox.Visible = false;
			}
			tableLayoutPanel2.Enabled = true;
		}
			
		public void Start()
		{
			// Initialise sampling parameters:
			device.SetSamplingFrequency(measurement.SamplingFrequency, measurement.Averages);
			GetSamplingParameters();
			SetPlotters();
			Reset(true);
			//
			startToolStripButton.Text = "Stop";
			//
			device.StartSampling();
			meterRefreshTimer.Start();
			//
			// Disable controls:
			aSensorControl.ReadOnly = true;
			aSensorEditButton.Enabled = false;
			bSensorControl.ReadOnly = true;
			bSensorEditButton.Enabled = false;
			cSensorControl.ReadOnly = true;
			cSensorEditButton.Enabled = false;
			rescanFTDIDevicesToolStripMenuItem.Enabled = false;
			loadMeasurementSetupFromFileToolStripMenuItem.Enabled = false;
			fsNumericControl.Enabled = false;
			dataBufferNumericControl.Enabled = false;
			averagesComboBox.Enabled = false;
			copyDataToClipboardToolStripMenuItem.Enabled = false;
			//
			aDataPlotter.Start();
			bDataPlotter.Start();
			cDataPlotter.Start();
			dataBuffer.Start();
			//
			isSampling = true;
			toolStripStatusLabel1.Text = "Sampling started.";
		}
		
		public void Stop()
		{
			startToolStripButton.Text = "Start";
			startToolStripButton.Enabled = false;
			toolStripStatusLabel1.Text = "Stopping...";
			this.Refresh();
			//
			meterRefreshTimer.Stop();
			aDataPlotter.Stop();
			bDataPlotter.Stop();
			cDataPlotter.Stop();
			//
			if (device != null) {
				device.StopSampling();
				device.SamplingStopped.WaitOne(10000);
			}
			dataBuffer.Stop();
			//
			isSampling = false;
			toolStripStatusLabel1.Text = "Sampling stopped.";
			// Enable controls:
			aSensorControl.ReadOnly = false;
			aSensorEditButton.Enabled = true;
			bSensorControl.ReadOnly = false;
			bSensorEditButton.Enabled = true;
			cSensorControl.ReadOnly = false;
			cSensorEditButton.Enabled = true;
			rescanFTDIDevicesToolStripMenuItem.Enabled = true;
			loadMeasurementSetupFromFileToolStripMenuItem.Enabled = true;
			fsNumericControl.Enabled = true;
			dataBufferNumericControl.Enabled = true;
			averagesComboBox.Enabled = true;
			copyDataToClipboardToolStripMenuItem.Enabled = true;
			//
			startToolStripButton.Enabled = true;
		}
		
		public void UpdateDataGridView(DataGridView dataGridView)
		{
			if (dataGridView.InvokeRequired) {
				MethodInvoker method = delegate { UpdateDataGridView(dataGridView); };
				dataGridView.Invoke(method);
			} else {
				LevelCrossingDetector detector = null;
				LevelCrossingCycleCollection cycles = null;
				DataGridViewRow row;
				int rowIndex;
				double length = 1.0;
				
				dataGridView.Rows.Clear();
				
				if (dataGridView == aDataGridView) {
					detector = aDataPlotter.LevelCrossingDetector;
					length = aLevelCrossingControl.Length;
				} else if (dataGridView == bDataGridView) {
					detector = bDataPlotter.LevelCrossingDetector;
					length = bLevelCrossingControl.Length;
				} else if (dataGridView == cDataGridView) {
					detector = cDataPlotter.LevelCrossingDetector;
					length = cLevelCrossingControl.Length;
				}
				
				if (detector == null) return;
				
				cycles = detector.Cycles;
				
				
				lock (cycles.SyncRoot) {
					for (int i = 0; i < cycles.Count; i++) {
						rowIndex = dataGridView.Rows.Add();
						row = dataGridView.Rows[rowIndex];
						row.Cells[0].Value = cycles[i].UpCrossing;
						if (!Double.IsNaN(cycles[i].NextUpCrossing)) {
							row.Cells[1].Value = cycles[i].Period;
						}
						if (!Double.IsNaN(cycles[i].DownCrossing)) {
							row.Cells[2].Value = length / cycles[i].PulseWidth;
						}
					}
				}
			}
		}
		#endregion
		
		#region EVENT HANDLERS
		protected void HandleButtonClickEvents(object sender, EventArgs e)
		{
			if (sender == startToolStripButton) {
				if (device == null) return;
				//
				if (isSampling) {
					Stop();
				} else {
					Start();
				}
			} else if (sender == aSensorEditButton) {
				using (SensorEditor editor = new SensorEditor(SensorEditorMode.Edit, (channelAInfo.Sensor != null) ? channelAInfo.Sensor.Clone() : null, device, 0)) {
					if (editor.ShowDialog() == DialogResult.OK) {
						channelAInfo.Sensor = editor.Sensor;
						aSensorControl.Sensor = channelAInfo.Sensor;
						measurement[0].Sensor = channelAInfo.Sensor;
						HandleSensorUpdatedEvents(aSensorControl, new EventArgs());
					}
				}
			} else if (sender == bSensorEditButton) {
				using (SensorEditor editor = new SensorEditor(SensorEditorMode.Edit, (channelBInfo.Sensor != null) ? channelBInfo.Sensor.Clone() : null, device, 1)) {
					if (editor.ShowDialog() == DialogResult.OK) {
						channelBInfo.Sensor = editor.Sensor;
						bSensorControl.Sensor = channelBInfo.Sensor;
						measurement[1].Sensor = channelBInfo.Sensor;
						HandleSensorUpdatedEvents(bSensorControl, new EventArgs());
					}
				}
			} else if (sender == cSensorEditButton) {
				using (SensorEditor editor = new SensorEditor(SensorEditorMode.Edit, (channelCInfo.Sensor != null) ? channelCInfo.Sensor.Clone() : null, device, 2)) {
					if (editor.ShowDialog() == DialogResult.OK) {
						channelCInfo.Sensor = editor.Sensor;
						cSensorControl.Sensor = channelCInfo.Sensor;
						measurement[2].Sensor = channelCInfo.Sensor;
						HandleSensorUpdatedEvents(cSensorControl, new EventArgs());
					}
				}
			} else if (sender == resetToolStripButton) {
				Reset(!isSampling);
			} else if (sender == showPanelToolStripButton) {
				if (splitContainer1.Panel2Collapsed == true) {
					splitContainer1.Panel2Collapsed = false;
					showPanelToolStripButton.Text = "Hide side panel";
				} else {
					splitContainer1.Panel2Collapsed = true;
					showPanelToolStripButton.Text = "Show side panel";
				}
				this.Refresh();
			} 
		}
		
		protected void HandleCheckBoxClickEvents(object sender, EventArgs e)
		{
			if (sender == aActiveCheckBox) {
				measurement[0].IsActive = aActiveCheckBox.Checked;
				SetChannelVisibility();
				SetTableVisibility();
				SetTextBoxes();
			} else if (sender == bActiveCheckBox) {
				measurement[1].IsActive = bActiveCheckBox.Checked;
				SetChannelVisibility();
				SetTableVisibility();
				SetTextBoxes();
			} else if (sender == cActiveCheckBox) {
				measurement[2].IsActive = cActiveCheckBox.Checked;
				SetChannelVisibility();
				SetTableVisibility();
				SetTextBoxes();
			} else if (sender == aAutoCheckBox) {
				if (aAutoCheckBox.Checked) {
					measurement[0].IsAutoscaled = true;
					// Min:
					aMinNumericControl.Enabled = false;
					aMinNumericControl.SetValue(Double.NaN, "Auto");
					aDataPlotter.ChartAreas[0].AxisY.Minimum = Double.NaN;
					aDataPlotter.ChartAreas[0].AxisY.IsStartedFromZero = false;
					// Max:
					aMaxNumericControl.Enabled = false;
					aMaxNumericControl.SetValue(Double.NaN, "Auto");
					aDataPlotter.ChartAreas[0].AxisY.Maximum = Double.NaN;
				} else {
					double yMin = aDataPlotter.ChartAreas[0].AxisY.Minimum;
					double yMax = aDataPlotter.ChartAreas[0].AxisY.Maximum;
					//
					measurement[0].IsAutoscaled = false;
					measurement[0].YMin = yMin;
					measurement[0].YMax = yMax;
					// Min:
					aMinNumericControl.Enabled = true;
					aMinNumericControl.SetValue(yMin);
					aDataPlotter.ChartAreas[0].AxisY.Minimum = yMin;
					// Max:
					aMaxNumericControl.Enabled = true;
					aMaxNumericControl.SetValue(yMax);
					aDataPlotter.ChartAreas[0].AxisY.Maximum = yMax;
				}
			} else if (sender == bAutoCheckBox) {
				if (bAutoCheckBox.Checked) {
					measurement[1].IsAutoscaled = true;
					// Min:
					bMinNumericControl.Enabled = false;
					bMinNumericControl.SetValue(Double.NaN, "Auto");
					bDataPlotter.ChartAreas[0].AxisY.Minimum = Double.NaN;
					bDataPlotter.ChartAreas[0].AxisY.IsStartedFromZero = false;
					// Max:
					bMaxNumericControl.Enabled = false;
					bMaxNumericControl.SetValue(Double.NaN, "Auto");
					bDataPlotter.ChartAreas[0].AxisY.Maximum = Double.NaN;
				} else {
					double yMin = bDataPlotter.ChartAreas[0].AxisY.Minimum;
					double yMax = bDataPlotter.ChartAreas[0].AxisY.Maximum;
					//
					measurement[1].IsAutoscaled = false;
					measurement[1].YMin = yMin;
					measurement[1].YMax = yMax;
					// Min:
					bMinNumericControl.Enabled = true;
					bMinNumericControl.SetValue(yMin);
					bDataPlotter.ChartAreas[0].AxisY.Minimum = yMin;
					// Max:
					bMaxNumericControl.Enabled = true;
					bMaxNumericControl.SetValue(yMax);
					bDataPlotter.ChartAreas[0].AxisY.Maximum = yMax;
				}
			} else if (sender == cAutoCheckBox) {
				if (cAutoCheckBox.Checked) {
					measurement[2].IsAutoscaled = true;
					// Min:
					cMinNumericControl.Enabled = false;
					cMinNumericControl.SetValue(Double.NaN, "Auto");
					cDataPlotter.ChartAreas[0].AxisY.Minimum = Double.NaN;
					cDataPlotter.ChartAreas[0].AxisY.IsStartedFromZero = false;
					// Max:
					cMaxNumericControl.Enabled = false;
					cMaxNumericControl.SetValue(Double.NaN, "Auto");
					cDataPlotter.ChartAreas[0].AxisY.Maximum = Double.NaN;
				} else {
					double yMin = cDataPlotter.ChartAreas[0].AxisY.Minimum;
					double yMax = cDataPlotter.ChartAreas[0].AxisY.Maximum;
					//
					measurement[2].IsAutoscaled = false;
					measurement[2].YMin = yMin;
					measurement[2].YMax = yMax;
					// Min:
					cMinNumericControl.Enabled = true;
					cMinNumericControl.SetValue(yMin);
					cDataPlotter.ChartAreas[0].AxisY.Minimum = yMin;
					// Max:
					cMaxNumericControl.Enabled = true;
					cMaxNumericControl.SetValue(yMax);
					cDataPlotter.ChartAreas[0].AxisY.Maximum = yMax;
				}
			}
		}
		
		protected void HandleColourChangedEvents(object sender, EventArgs e)
		{
			if (sender == aColourPickerControl) {
				measurement[0].Colour = aColourPickerControl.Colour;
				aTextBox.ForeColor = measurement[0].Colour;
				aDataPlotter.Series[0].Color = measurement[0].Colour;
			} else if (sender == bColourPickerControl) {
				measurement[1].Colour = bColourPickerControl.Colour;
				bTextBox.ForeColor = measurement[1].Colour;
				bDataPlotter.Series[0].Color = measurement[1].Colour;
			} else if (sender == cColourPickerControl) {
				measurement[2].Colour = cColourPickerControl.Colour;
				cTextBox.ForeColor = measurement[2].Colour;
				cDataPlotter.Series[0].Color = measurement[2].Colour;
			}
		}
		
		protected void HandleContextMenuItemClickEvents(object sender, EventArgs e)
		{
			if (sender == copyContentToClipboardToolStripMenuItem) {
				DataGridView dataGridview = tableContextMenuStrip.SourceControl as DataGridView;
				
				CopyToClipBoard(dataGridview);
			} else if (sender == clearToolStripMenuItem) {
				DataGridView dataGridView = tableContextMenuStrip.SourceControl as DataGridView;
				if (dataGridView == aDataGridView) {
					if (aDataPlotter.LevelCrossingDetector != null) {
						lock (aDataPlotter.LevelCrossingDetector.Cycles.SyncRoot) {
							aDataPlotter.LevelCrossingDetector.Cycles.Clear();
						}
					}
				} else if (dataGridView == bDataGridView) {
					if (bDataPlotter.LevelCrossingDetector != null) {
						lock (bDataPlotter.LevelCrossingDetector.Cycles.SyncRoot) {
							bDataPlotter.LevelCrossingDetector.Cycles.Clear();
						}
					}
				} else if (dataGridView == cDataGridView) {
					if (cDataPlotter.LevelCrossingDetector != null) {
						lock (cDataPlotter.LevelCrossingDetector.Cycles.SyncRoot) {
							cDataPlotter.LevelCrossingDetector.Cycles.Clear();
						}
					}
				}
				
				UpdateDataGridView(dataGridView);
			} else if (sender == saveAsImageToolStripMenuItem) {
				MSDataPlotter plotter = plotterContextMenuStrip.SourceControl as MSDataPlotter;
				
				using (SaveFileDialog saveDialogue = new SaveFileDialog()) {
					saveDialogue.Filter = "Portable network graphics format (*.png)|*.png| Jpeg format (*.jpg)|*.jpg";
					if (saveDialogue.ShowDialog() == DialogResult.OK) {
						ChartImageFormat format;
						switch (saveDialogue.FilterIndex) {
							case 2:
								format = ChartImageFormat.Jpeg;
								break;
							case 1:
							default:
								format = ChartImageFormat.Png;
								break;
						}
						plotter.SaveImage(saveDialogue.FileName, format);
					}
				}
			} else if (sender == copyDataToClipboardToolStripMenuItem) {
				MSDataPlotter plotter = plotterContextMenuStrip.SourceControl as MSDataPlotter;
				
				plotter.CopyToClipBoard();
			} else if (sender == resetToolStripMenuItem) {
				Reset(!isSampling);
			}
		}
		
		protected void HandleCyclesUpdatedEvents(object sender, EventArgs e)
		{
			if (sender == aDataPlotter.LevelCrossingDetector) {
				UpdateDataGridView(aDataGridView);
			} else if (sender == bDataPlotter.LevelCrossingDetector) {
				UpdateDataGridView(bDataGridView);
			} else if (sender == cDataPlotter.LevelCrossingDetector) {
				UpdateDataGridView(cDataGridView);
			}
		}
		
		protected void HandleLevelCrossingControlUpdatedEvents(object sender, EventArgs e)
		{
			if (sender == aLevelCrossingControl) {
				if (aDataPlotter.LevelCrossingDetector == null) {
					aDataPlotter.LevelCrossingDetector = aLevelCrossingControl.CreateLevelCrossingDetector();
					aDataPlotter.LevelCrossingDetector.Cycles.Capacity = defaultCycles;
					aDataPlotter.LevelCrossingDetector.CyclesUpdated += HandleCyclesUpdatedEvents;
				} else {
					aLevelCrossingControl.SetLevelCrossingDetector(aDataPlotter.LevelCrossingDetector);
				}
				measurement[0].LevelCrossingDetector = aDataPlotter.LevelCrossingDetector;
				measurement[0].ObjectLength = aLevelCrossingControl.Length;
			} else if (sender == bLevelCrossingControl) {
				if (bDataPlotter.LevelCrossingDetector == null) {
					bDataPlotter.LevelCrossingDetector = bLevelCrossingControl.CreateLevelCrossingDetector();
					bDataPlotter.LevelCrossingDetector.Cycles.Capacity = defaultCycles;
					bDataPlotter.LevelCrossingDetector.CyclesUpdated += HandleCyclesUpdatedEvents;
				} else {
					bLevelCrossingControl.SetLevelCrossingDetector(bDataPlotter.LevelCrossingDetector);
				}
				measurement[1].LevelCrossingDetector = bDataPlotter.LevelCrossingDetector;
				measurement[1].ObjectLength = bLevelCrossingControl.Length;
			} else if (sender == cLevelCrossingControl) {
				if (cDataPlotter.LevelCrossingDetector == null) {
					cDataPlotter.LevelCrossingDetector = cLevelCrossingControl.CreateLevelCrossingDetector();
					cDataPlotter.LevelCrossingDetector.Cycles.Capacity = defaultCycles;
					cDataPlotter.LevelCrossingDetector.CyclesUpdated += HandleCyclesUpdatedEvents;
				} else {
					cLevelCrossingControl.SetLevelCrossingDetector(cDataPlotter.LevelCrossingDetector);
				}
				measurement[2].LevelCrossingDetector = cDataPlotter.LevelCrossingDetector;
				measurement[2].ObjectLength = cLevelCrossingControl.Length;
			} else {
				return;
			}
			
			SetTableVisibility();
		}
		
		protected void HandleMenuItemClickEvents(object sender, EventArgs e)
		{
			if (sender == rescanFTDIDevicesToolStripMenuItem) {
				SetFtdi();
				InitialiseDevice();
			}  else if (sender == fromChannelAToolStripMenuItem) {
				using (SaveFileDialog dialogue = new SaveFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					if (channelAInfo.Sensor != null) dialogue.FileName = channelAInfo.Sensor.Name;
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						if (channelAInfo.Sensor != null) {
							channelAInfo.Sensor.Save(dialogue.FileName);
						}
					}
				}
			} else if (sender == fromChannelBToolStripMenuItem) {
				using (SaveFileDialog dialogue = new SaveFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					if (channelBInfo.Sensor != null) dialogue.FileName = channelBInfo.Sensor.Name;
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						if (channelBInfo.Sensor != null) {
							channelBInfo.Sensor.Save(dialogue.FileName);
						}
					}
				}
			} else if (sender == fromChannelCToolStripMenuItem) {
				using (SaveFileDialog dialogue = new SaveFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					if (channelCInfo.Sensor != null) dialogue.FileName = channelCInfo.Sensor.Name;
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						if (channelCInfo.Sensor != null) {
							channelCInfo.Sensor.Save(dialogue.FileName);
						}
					}
				}
			} else if (sender == toChannelAToolStripMenuItem) {
				using (OpenFileDialog dialogue = new OpenFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						channelAInfo.Sensor = Sensor.Open(dialogue.FileName);
						aSensorControl.Sensor = channelAInfo.Sensor;
						measurement[0].Sensor = channelAInfo.Sensor;
					}
				}
			} else if (sender == toChannelBToolStripMenuItem) {
				using (OpenFileDialog dialogue = new OpenFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						channelBInfo.Sensor = Sensor.Open(dialogue.FileName);
						bSensorControl.Sensor = channelBInfo.Sensor;
						measurement[1].Sensor = channelBInfo.Sensor;
					}
				}
			} else if (sender == toChannelCToolStripMenuItem) {
				using (OpenFileDialog dialogue = new OpenFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						channelCInfo.Sensor = Sensor.Open(dialogue.FileName);
						cSensorControl.Sensor = channelCInfo.Sensor;
						measurement[2].Sensor = channelCInfo.Sensor;
					}
				}
			} else if (sender == saveMeasurementSetupToolStripMenuItem) {
				using (SaveFileDialog dialogue = new SaveFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						if (measurement != null) measurement.Save(dialogue.FileName);
					}
				}
			} else if (sender == loadMeasurementSetupFromFileToolStripMenuItem) {
				using (OpenFileDialog dialogue = new OpenFileDialog()) {
					dialogue.Filter = "Xml files (*.xml)|*.xml";
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						LoadMeasurement(dialogue.FileName);
					}
				}
			} else if (sender == saveMeasurementDataToolStripMenuItem) {
				using (SaveFileDialog dialogue = new SaveFileDialog()) {
					dialogue.SupportMultiDottedExtensions = true;
					dialogue.Filter = "Text files (*.data.txt)|*.data.txt";
					
					if (dialogue.ShowDialog() == DialogResult.OK) {
						if (dataBuffer != null) {
							string header = String.Format("// Time increment: {0} s", 1.0 / measurement.SamplingFrequency);
							
							header += Environment.NewLine +
								String.Format("// Channel {0}: {1}", Edaq530.ChannelNames[0], (channelAInfo.Sensor == null) ? "U [V]" :
								              channelAInfo.Sensor.GetLabel());
							header += Environment.NewLine +
								String.Format("// Channel {0}: {1}", Edaq530.ChannelNames[1], (channelBInfo.Sensor == null) ? "U [V]" :
								              channelBInfo.Sensor.GetLabel());
							header += Environment.NewLine +
								String.Format("// Channel {0}: {1}", Edaq530.ChannelNames[2], (channelCInfo.Sensor == null) ? "U [V]" :
								              channelCInfo.Sensor.GetLabel());
							
							dataBuffer.Save(dialogue.FileName, header, "f4", 1.0 / measurement.SamplingFrequency);
						}
					}
				}
			}
		}
		
		protected void HandleValueChangedEvents(object sender, EventArgs e)
		{
			if (sender == fsNumericControl) {
				if (isSampling) return;
				if (!device.SamplingStopped.WaitOne(0)) return;
				if (!Double.IsNaN(fsNumericControl.Value)) {
					device.SetSamplingFrequency(fsNumericControl.Value, measurement.Averages);
					GetSamplingParameters();
					SetPlotters();
				}
			} else if (sender == aMinNumericControl) {	
				if (Double.IsNaN(aDataPlotter.ChartAreas[0].AxisY.Maximum) ||
					    (aMinNumericControl.Value < aDataPlotter.ChartAreas[0].AxisY.Maximum)) {
					aDataPlotter.ChartAreas[0].AxisY.Minimum = aMinNumericControl.Value;
					measurement[0].YMin = aMinNumericControl.Value;
				} else {
					aMinNumericControl.SetValue(aDataPlotter.ChartAreas[0].AxisY.Minimum, "Auto");
				}
			} else if (sender == aMaxNumericControl) {
				if (aMaxNumericControl.Value > aDataPlotter.ChartAreas[0].AxisY.Minimum) {
					aDataPlotter.ChartAreas[0].AxisY.Maximum = aMaxNumericControl.Value;
					measurement[0].YMax = aMaxNumericControl.Value;
				} else {
					aMaxNumericControl.SetValue(aDataPlotter.ChartAreas[0].AxisY.Maximum, "Auto");
				}
			} else if (sender == bMinNumericControl) {	
				if (Double.IsNaN(bDataPlotter.ChartAreas[0].AxisY.Maximum) ||
					    (bMinNumericControl.Value < bDataPlotter.ChartAreas[0].AxisY.Maximum)) {
					bDataPlotter.ChartAreas[0].AxisY.Minimum = bMinNumericControl.Value;
					measurement[1].YMin = bMinNumericControl.Value;
				} else {
					bMinNumericControl.SetValue(bDataPlotter.ChartAreas[0].AxisY.Minimum, "Auto");
				}
			} else if (sender == bMaxNumericControl) {
				if (bMaxNumericControl.Value > bDataPlotter.ChartAreas[0].AxisY.Minimum) {
					bDataPlotter.ChartAreas[0].AxisY.Maximum = bMaxNumericControl.Value;
					measurement[1].YMax = bMaxNumericControl.Value;
				} else {
					bMaxNumericControl.SetValue(bDataPlotter.ChartAreas[0].AxisY.Maximum, "Auto");
				}
			} else if (sender == cMinNumericControl) {	
				if (Double.IsNaN(cDataPlotter.ChartAreas[0].AxisY.Maximum) ||
					    (cMinNumericControl.Value < cDataPlotter.ChartAreas[0].AxisY.Maximum)) {
					cDataPlotter.ChartAreas[0].AxisY.Minimum = cMinNumericControl.Value;
					measurement[2].YMin = cMinNumericControl.Value;
				} else {
					cMinNumericControl.SetValue(cDataPlotter.ChartAreas[0].AxisY.Minimum, "Auto");
				}
			} else if (sender == cMaxNumericControl) {
				if (cMaxNumericControl.Value > cDataPlotter.ChartAreas[0].AxisY.Minimum) {
					cDataPlotter.ChartAreas[0].AxisY.Maximum = cMaxNumericControl.Value;
					measurement[2].YMax = cMaxNumericControl.Value;
				} else {
					cMaxNumericControl.SetValue(cDataPlotter.ChartAreas[0].AxisY.Maximum, "Auto");
				}
			} else if (sender == xScaleTitleLabelledTextBox) {
				aDataPlotter.ChartAreas[0].AxisX.Title = xScaleTitleLabelledTextBox.TextBox.Text;
				bDataPlotter.ChartAreas[0].AxisX.Title = xScaleTitleLabelledTextBox.TextBox.Text;
				cDataPlotter.ChartAreas[0].AxisX.Title = xScaleTitleLabelledTextBox.TextBox.Text;
				measurement.XScaleTitle = xScaleTitleLabelledTextBox.TextBox.Text;
			} else if (sender == xWidthNumericControl) {
				if (xWidthNumericControl.Value == 0.0) {
					xWidthNumericControl.TextBox.Text = String.Empty;
					return;
				}
				double dt = 1.0 / measurement.SamplingFrequency;
				int points = (int)Math.Round(xWidthNumericControl.Value / dt, MidpointRounding.AwayFromZero);
				if (points > maxPoints) points = maxPoints;
				xPointsNumericControl.SetValue(points);
				xWidthNumericControl.SetValue(points * dt);
				measurement.XScalePoints = points;
				SetPlotters();
			} else if (sender == xPointsNumericControl) {
				if (xPointsNumericControl.Value < 2.0) {
					xPointsNumericControl.TextBox.Text = String.Empty;
					return;
				}
				double dt = 1.0 / measurement.SamplingFrequency;
				int points = (int)xPointsNumericControl.Value;
				xWidthNumericControl.SetValue(points * dt);
				measurement.XScalePoints = points;
				SetPlotters();
			} else if (sender == dataBufferNumericControl) {
				dataBuffer.Capacity = (int)dataBufferNumericControl.Value;
				measurement.DataBufferSize = dataBuffer.Capacity;
			} else if (sender == refreshRateNumericControl) {
				measurement.RefreshInterval = (int)Math.Round(1000.0 / refreshRateNumericControl.Value, MidpointRounding.AwayFromZero);
				aDataPlotter.Interval = measurement.RefreshInterval;
				bDataPlotter.Interval = measurement.RefreshInterval;
				cDataPlotter.Interval = measurement.RefreshInterval;
			}
		}
		
		protected void HandlePanelExpandingEvents(object sender, XPanderStateChangeEventArgs e)
		{
			xPanderPanelList1.Refresh();
			
			if (sender is XPanderPanel) {
				XPanderPanel panel = sender as XPanderPanel;
				panel.Refresh();
			}
		}
	
		protected void HandleSelectionChangeCommittedEvents(object sender, EventArgs e)
		{
			if (sender == averagesComboBox) {
				device.SetSamplingFrequency(measurement.SamplingFrequency, Averages.FromCode((byte)averagesComboBox.SelectedIndex));
				GetSamplingParameters();
			}
		}
		
		protected void HandleSensorRecreatedEvents(object sender, EventArgs e)
		{
			if (sender == aSensorControl) {
				measurement[0].Sensor = aSensorControl.Sensor;
				channelAInfo.Sensor = aSensorControl.Sensor;
				HandleSensorUpdatedEvents(sender, e);
			} else if (sender == bSensorControl) {
				measurement[1].Sensor = bSensorControl.Sensor;
				channelBInfo.Sensor = bSensorControl.Sensor;
				HandleSensorUpdatedEvents(sender, e);
			} else if (sender == cSensorControl) {
				measurement[2].Sensor = cSensorControl.Sensor;
				channelCInfo.Sensor = cSensorControl.Sensor;
				HandleSensorUpdatedEvents(sender, e);
			}
		}
		
		protected void HandleSensorUpdatedEvents(object sender, EventArgs e)
		{
			if (sender == aSensorControl) {
				if (channelAInfo.Sensor == null) {
					aDataPlotter.ChartAreas[0].AxisY.Title = "U [V]";
				} else {
					aDataPlotter.ChartAreas[0].AxisY.Title = channelAInfo.Sensor.GetLabel();
				}
			} else if (sender == bSensorControl) {
				if (channelBInfo.Sensor == null) {
					bDataPlotter.ChartAreas[0].AxisY.Title = "U [V]";
				} else {
					bDataPlotter.ChartAreas[0].AxisY.Title = channelBInfo.Sensor.GetLabel();
				}
			} else if (sender == cSensorControl) {
				if (channelCInfo.Sensor == null) {
					cDataPlotter.ChartAreas[0].AxisY.Title = "U [V]";
				} else {
					cDataPlotter.ChartAreas[0].AxisY.Title = channelCInfo.Sensor.GetLabel();
				}
			}
		}
		
		protected void HandleSplitterMovedEvents(object sender, SplitterEventArgs e)
		{
			this.Refresh();
		}
		#endregion
		
		#region OVERRIDES
		[System.ComponentModel.EditorBrowsableAttribute()]
		protected override void OnResize(EventArgs e)
		{
			base.OnResize(e);
			SetTextBoxes();
			this.Refresh();
		}
		
		[System.ComponentModel.EditorBrowsableAttribute()]
		protected override void OnFormClosing(FormClosingEventArgs e)
		{
			if (isSampling) Stop();
			if ((ftdi != null) && ftdi.IsOpen) ftdi.Close();
			measurement.Save("LastSettings.xml");
			base.OnFormClosing(e);
		}
		#endregion
	}
}
