﻿/*
 * Készítette a SharpDevelop.
 * Felhasználó: Ouroboros
 * Dátum: 2010.02.28.
 * Idő: 13:01
 * 
 * A sablon megváltoztatásához használja az Eszközök | Beállítások | Kódolás | Szabvány Fejlécek Szerkesztését.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Text;
using System.Threading;
//
using TetheredSun.Core;

namespace Noise.DataAcquisition
{
	/// <summary>
	/// Description of DataBuffer.
	/// </summary>
	public class DataBuffer
	{
		public static int DefaultCapacity = 65536;
		//
		protected int dueTime = Timeout.Infinite;
		protected int interval = 100;
		protected string separator = "\t";
		protected ChannelDataSocketCollection channels;
		protected System.Threading.Timer timer;
		protected volatile bool isReady = true;
		protected volatile bool stopped = false;
		protected RingBuffer<double[]> buffer;
		
		
		public DataBuffer(int capacity)
		{
			this.channels = new ChannelDataSocketCollection();
			this.buffer = new RingBuffer<double[]>(capacity);
			this.timer = new System.Threading.Timer(new TimerCallback(UpdateBuffer), null, Timeout.Infinite, interval);
		}
		
		public DataBuffer() : this(DefaultCapacity) { }
		
		
		public ChannelDataSocketCollection Channels {
			get { return channels; }
		}
		
		public int Count {
			get { lock (buffer.SyncRoot) { return buffer.Count; } }
		}
		
		public int Capacity {
			get { lock (buffer.SyncRoot) { return buffer.Capacity; } }
			set { lock (buffer.SyncRoot) { buffer.Capacity = value; } }
		}
		
		public int Interval {
			get { return interval; }
			set {
				interval = value;
				timer.Change(dueTime, this.interval);
			}
		}
		
		public string Separator {
			get { return separator; }
			set { separator = value; }
		}
		
		public void Clear()
		{
			lock (buffer.SyncRoot) { buffer.Clear(); }
		}
		
		public void Save(string path, string header, string format, double dt)
		{
			int bufferSize;
			double[] item;
			string line;
			
			if (String.IsNullOrEmpty(format)) format = "f3";
			
			lock (buffer.SyncRoot) {
				bufferSize = buffer.Count;
				if (bufferSize > 0) {
					using (StreamWriter writer = new StreamWriter(path, false, Encoding.UTF8, bufferSize)) {
						if (!String.IsNullOrEmpty(header)) writer.WriteLine(header);
						
						for (int i = 0; i < bufferSize; i++) {
							item = buffer[i];
							line = (i * dt).ToString(format) + separator;
							
							for (int j = 0; j < item.Length; j++) {
								line += item[j].ToString(format);
								if (j < item.Length - 1) line += separator;
							}
							
							writer.WriteLine(line);
						}
					}
				}
			}
		}
		
		public void Start()
		{
			stopped = false;
			dueTime = 0;
			timer.Change(dueTime, interval);
		}
		
		public void Start(int interval)
		{
			this.interval = interval;
			Start();
		}
		
		public void Stop()
		{
			stopped = true;
			dueTime = Timeout.Infinite;
			timer.Change(dueTime, interval);
		}
	
	
		protected void UpdateBuffer(object state)
		{
			if (!isReady) return;
			if (stopped) return;
			
			int dataPerChannel;
			double[] item = null;
			
			lock (channels.SyncRoot) { 
				dataPerChannel = channels.GetChannelDataCount();
				
				for (int i = 0; i < dataPerChannel; i++) {
					item = channels.Dequeue();
					if (item != null) {
						lock (buffer.SyncRoot) {
							buffer.Add(item);
						}
					}
				}
			}
		}
	}
}
