﻿/*
 * Készítette a SharpDevelop.
 * Felhasználó: phil
 * Dátum: 2009.07.06.
 * Idő: 14:22
 * 
 * A sablon megváltoztatásához használja az Eszközök | Beállítások | Kódolás | Szabvány Fejlécek Szerkesztését.
 */
using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Windows.Forms;

namespace Noise.DataAcquisition
{
	public enum LogType {
		Binary,
		Text
	}
	
	/// <summary>
	/// Description of DataLogger.
	/// </summary>
	public class DataLogger
	{
		protected string numberFormat = "f3";
		protected string path;
		protected string separator = "\t";
		protected Timer timer;
		protected IChannelDataSocket<double[]> data;
		protected volatile bool recordTime = false;
		protected volatile bool isReady = true;
		protected double? startTime = null;
		protected double? timeIncrement = null;
		protected double? time;
		protected LogType logType = LogType.Text;
		
		
		
		public DataLogger(string path, IChannelDataSocket<double[]> data)
		{
			this.path = path;
			this.data = data;
			timer = new Timer();
			timer.Interval = 1000;
			timer.Tick += delegate(object sender, EventArgs e) { TextLog(); };
		}
		
		public DataLogger(IChannelDataSocket<double[]> data) : this(null, data) { }
		
		
		
		public string NumberFormat {
			get { return numberFormat; }
			set { numberFormat = value; }
		}
		
		public string Path {
			get { return path; }
			set { path = value; }
		}
		
		public string Separator {
			get { return separator; }
			set { separator = value; }
		}
		
		public Timer Timer {
			get { return timer; }
		}
		
		public IChannelDataSocket<double[]> Data {
			get { return data; }
		}
		
		public LogType LogType {
			get { return logType; }
			set {
				logType = value;
				switch (logType) {
					case LogType.Text:
					default:
						int interval = timer.Interval;
						timer = new Timer();
						timer.Interval = interval;
						timer.Tick += delegate(object sender, EventArgs e) { TextLog(); };
						break;
				}
			}
		}
		
		public bool RecordTime {
			get { return recordTime; }
			set { recordTime = value; }
		}
		
		public Nullable<double> StartTime {
			get { return startTime; }
			set {
				startTime = value;
				ResetTime();
			}
		}
		
		public Nullable<double> TimeIncrement {
			get { return timeIncrement; }
			set { timeIncrement = value; }
		}
		
		public Nullable<double> Time {
			get { return time; }
		}
				
		
		
		public void BrowsePath()
		{
			using (SaveFileDialog dialogue = new SaveFileDialog()) {
				if (dialogue.ShowDialog() == DialogResult.OK) {
					path = dialogue.FileName;
				}
			}
		}
		
		public void ResetTime()
		{
			time = startTime;
		}
		
		
		
		protected void TextLog()
		{
			if (!isReady) return;
			if (String.IsNullOrEmpty(path)) return;
			isReady = false;
			//
			StreamWriter writer = null;
			double[] item;
			bool continueLogging = true;
			
			try {
				if (data.Count > 0) {
					writer = new StreamWriter(path, true);
					while (continueLogging) {
						lock ((data as ICollection).SyncRoot) {
							item = data.Dequeue();
							continueLogging = (data.Count > 0);
						}
						
						if (recordTime) {
							writer.Write(time.Value.ToString(numberFormat) + separator);
							time += timeIncrement.Value;
						}
						
						writer.Write(item[0].ToString(numberFormat));
						
						for (int i = 1; i < item.Length; i++) {
							writer.Write(separator + item[i].ToString(numberFormat));
						}
						
						writer.WriteLine();
					}
				}
			} catch (Exception exception) {
				throw exception;
			} finally {
				if (writer != null) {
					writer.Close();
				}
				isReady = true;
			}
		}
	}
}
